###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import List, Optional, Union

from everysk.api import Datastore
from everysk.api.utils import EveryskList
from mcp.server.fastmcp import FastMCP

from everysk_mcp.fields import definitions as fields


###############################################################################
# Implementation
###############################################################################
def register_datastore_tools(mcp: FastMCP) -> None:
    """Register all datastore tools with the given MCP instance."""

    @mcp.tool(
        description="Returns a list of datastores previously created. The datastores are returned in sorted order, with the most recent datastore appearing first."
    )
    def datastore_list(
        query: Optional[str] = fields.query,
        workspace: Optional[str] = fields.workspace,
        page_size: Optional[int] = fields.page_size,
    ) -> EveryskList[Datastore]:
        result = Datastore.list(query=query, workspace=workspace, page_size=page_size)
        return result

    @mcp.tool(
        description="Retrieves the details of an existing datastore by supplying the datastore's id."
    )
    def datastore_retrieve(
        id: str = fields.id_, workspace: Optional[str] = fields.workspace
    ) -> Datastore:
        result = Datastore.retrieve(id=id, workspace=workspace)
        return result

    @mcp.tool(
        description="Filters a list of datastores based on the provided criteria."
    )
    def datastore_filter(
        workspace: Optional[str] = fields.workspace,
        limit: Optional[int] = fields.limit,
        tags: Optional[List[str]] = fields.tags,
        start: Optional[str] = fields.start_date,
        end: Optional[str] = fields.end_date,
        link_uid: Optional[str] = fields.link_uid,
    ) -> EveryskList[Datastore]:
        result = Datastore.filter(
            workspace=workspace,
            limit=limit,
            tags=tags,
            start=start,
            end=end,
            link_uid=link_uid,
        )
        return result

    @mcp.tool(
        description="Creates a new datastore and then returns the created datastore."
    )
    def datastore_create(
        name: str = fields.name,
        description: Optional[str] = fields.description,
        tags: List[str] = fields.tags,
        date: Optional[str] = fields.date,
        data: List[List[Union[str, int, float, bool]]] = fields.data_datastore,
        with_data: Optional[bool] = fields.with_data,
        workspace: Optional[str] = fields.workspace,
    ) -> Datastore:
        result = Datastore.create(
            name=name,
            description=description,
            tags=tags,
            date=date,
            data=data,
            with_data=with_data,
            workspace=workspace,
        )

        return result



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import List, Optional

from everysk.api import Portfolio
from everysk.api.utils import EveryskList
from mcp.server.fastmcp import FastMCP

from everysk_mcp.blueprints.securities import Security
from everysk_mcp.fields import definitions as fields


###############################################################################
# Implementation
###############################################################################
def register_portfolio_tools(mcp: FastMCP) -> None:
    """Register all portfolio tools with the given MCP instance."""

    @mcp.tool(
        description="Returns a list of portfolios previously created. The portfolios are returned in sorted order, with the most recent portfolio appearing first."
    )
    def portfolio_list(
        query: Optional[str] = fields.query,
        workspace: Optional[str] = fields.workspace,
        page_size: Optional[int] = fields.page_size,
    ) -> EveryskList[Portfolio]:
        result = Portfolio.list(query=query, workspace=workspace, page_size=page_size)
        return result

    @mcp.tool(
        description="Retrieves the details of an existing portfolio by supplying the portfolio's id."
    )
    def portfolio_retrieve(
        id: str = fields.id_, workspace: Optional[str] = fields.workspace
    ) -> Portfolio:
        result = Portfolio.retrieve(id=id, workspace=workspace)
        return result

    @mcp.tool(
        description="Filters a list of portfolios based on the provided criteria."
    )
    def portfolio_filter(
        workspace: Optional[str] = fields.workspace,
        limit: Optional[int] = fields.limit,
        tags: Optional[List[str]] = fields.tags,
        start: Optional[str] = fields.start_date,
        end: Optional[str] = fields.end_date,
        link_uid: Optional[str] = fields.link_uid,
    ) -> EveryskList[Portfolio]:
        result = Portfolio.filter(
            workspace=workspace,
            limit=limit,
            tags=tags,
            start=start,
            end=end,
            link_uid=link_uid,
        )
        return result

    @mcp.tool(
        description="Creates a new portfolio and then returns the created portfolio."
    )
    def portfolio_create(
        name: str = fields.name,
        description: Optional[str] = fields.description,
        tags: List[str] = fields.tags,
        date: Optional[str] = fields.date_portfolio,
        base_currency: Optional[str] = fields.base_currency,
        nlv: Optional[float] = fields.nlv,
        securities: List[Security] = fields.security_list,
        with_securities: Optional[bool] = fields.with_securities,
        workspace: Optional[str] = fields.workspace,
    ) -> Portfolio:
        if not securities:
            raise ValueError("At least one security must be provided in the portfolio.")

        if not isinstance(securities, list):
            raise TypeError("Securities must be a list of Security objects.")

        for security in securities:
            if not isinstance(security, Security):
                raise TypeError(
                    "Each security must be an instance of the Security class."
                )

        securities = [security.model_dump() for security in securities]

        result = Portfolio.create(
            name=name,
            description=description,
            tags=tags,
            date=date,
            base_currency=base_currency,
            nlv=nlv,
            securities=securities,
            with_securities=with_securities,
            workspace=workspace,
        )

        return result



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ["main", "server"]
from . import server


def main() -> None:
    server.mcp.run(transport="stdio")



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Annotated, Literal, Optional, Union

from everysk.core.datetime.date import Date
from pydantic import BaseModel, Field

from everysk_mcp.fields import securities as fields


###############################################################################
# Implementation
###############################################################################
class BrazilianFutures(BaseModel):
    type: Literal["brazilian_futures"] = Field(default="brazilian_futures", frozen=True)
    ticker: str = fields.ticker_br
    expiration_date: str = fields.expiration_contract
    price: str = fields.price

    def __str__(self) -> str:
        return f"BRFUT:{self.ticker} {self.expiration_date} {self.price}"


class BrazilianFixedIncome(BaseModel):
    type: Literal["brazilian_fixed_income"] = Field(
        default="brazilian_fixed_income", frozen=True
    )
    asset_code: str = fields.asset_code
    price: str = fields.price

    def __str__(self) -> str:
        return f"BRRF:{self.asset_code} {self.price}"

    @property
    def repo_agreement(self) -> str:
        next_business_day = Date.today().nearest_business_day().strftime("%Y%m%d")
        return f"REPO:{self.asset_code} {next_business_day} {self.price}"


class USEquitiesETFsMutualFunds(BaseModel):
    type: Literal["us_equities_etfs_mutual_funds"] = Field(
        default="us_equities_etfs_mutual_funds", frozen=True
    )
    ticker: str = fields.ticker_us

    def __str__(self) -> str:
        return self.ticker


class ForeignEquities(BaseModel):
    type: Literal["foreign_equities"] = Field(default="foreign_equities", frozen=True)
    ticker: str = fields.ticker_foreign_equities
    mic: str = fields.mic

    def __str__(self) -> str:
        return f"{self.ticker}:{self.mic}"


class EquityOptions(BaseModel):
    type: Literal["equity_options"] = Field(default="equity_options", frozen=True)
    ticker: str = fields.ticker_equity_options
    expiration_date: str = fields.expiration_contract
    call_or_put: Literal["C", "P"] = fields.call_or_put
    strike: str = fields.strike_options
    premium: str = fields.premium

    def __str__(self) -> str:
        return f"{self.ticker} {self.expiration_date} {self.call_or_put}{self.strike} {self.premium}"


class FXForward(BaseModel):
    type: Literal["fx_forward"] = Field(default="fx_forward", frozen=True)
    cross_currency: str = fields.cc_fx_forward
    expiration_date: str = fields.expiration_contract
    forward_rate: str = fields.forward_rate

    def __str__(self) -> str:
        return f"{self.cross_currency} {self.expiration_date} {self.forward_rate}"


class FXOptions(BaseModel):
    type: Literal["fx_options"] = Field(default="fx_options", frozen=True)
    cross_currency: str = fields.cc_fx_options
    expiration_date: str = fields.expiration_contract
    forward_rate: str = fields.forward_rate
    call_or_put: Literal["C", "P"] = fields.call_or_put
    strike: str = fields.strike_options
    premium: str = fields.premium

    def __str__(self) -> str:
        return f"{self.cross_currency} {self.expiration_date} {self.forward_rate} {self.call_or_put}{self.strike} {self.premium}"


class Futures(BaseModel):
    type: Literal["futures"] = Field(default="futures", frozen=True)
    contract: str = fields.contract
    expiration_date: str = fields.expiration_contract

    def __str__(self) -> str:
        return f"FUT:{self.contract} {self.expiration_date}"


class FutureOptions(BaseModel):
    type: Literal["future_options"] = Field(default="future_options", frozen=True)
    contract: str = fields.contract
    expiration_date: str = fields.expiration_contract
    call_or_put: Literal["C", "P"] = fields.call_or_put
    strike: str = fields.strike_options
    premium: str = fields.premium

    def __str__(self) -> str:
        return f"FUT: {self.contract} {self.expiration_date} {self.call_or_put}{self.strike} {self.premium}"


class CorporateBonds(BaseModel):
    type: Literal["corporate_bonds"] = Field(default="corporate_bonds", frozen=True)
    ticker: str = fields.ticker_bonds
    maturity_date: str = fields.maturity
    coupon: str = fields.coupon_bonds
    price: str = fields.price_bonds
    rating: Optional[str] = fields.rating

    def __str__(self) -> str:
        out = f"CORP:{self.ticker} {self.maturity_date} {self.coupon} {self.price}"
        if self.rating:
            out += f" {self.rating}"
        return out


class MunicipalBonds(BaseModel):
    type: Literal["municipal_bonds"] = Field(default="municipal_bonds", frozen=True)
    issuer: str = fields.issuer_muni_bonds
    maturity_date: str = fields.maturity
    coupon: str = fields.coupon_bonds
    price: str = fields.price_bonds
    rating: Optional[str] = fields.rating

    def __str__(self) -> str:
        out = f"MUNI:{self.issuer} {self.maturity_date} {self.coupon} {self.price}"
        if self.rating:
            out += f" {self.rating}"

        return out


class ConvertibleBonds(BaseModel):
    type: Literal["convertible_bonds"] = Field(default="convertible_bonds", frozen=True)
    ticker: str = fields.ticker_bonds
    maturity_date: str = fields.maturity
    coupon: str = fields.coupon_bonds
    price: str = fields.price_bonds
    strike: str = fields.strike
    rating: Optional[str] = fields.rating

    def __str__(self) -> str:
        out = f"CONV:{self.ticker} {self.maturity_date} {self.coupon} {self.price} {self.strike}"
        if self.rating:
            out += f" {self.rating}"
        return out


class ConvertiblePreferredBonds(BaseModel):
    type: Literal["convertible_preferred_bonds"] = Field(
        default="convertible_preferred_bonds", frozen=True
    )
    ticker: str = fields.ticker_bonds
    maturity_date: str = fields.maturity
    coupon: str = fields.coupon_bonds
    par: str = fields.par
    price: str = fields.price_bonds
    strike: str = fields.strike
    rating: Optional[str] = fields.rating

    def __str__(self) -> str:
        out = f"CONVP:{self.ticker} {self.maturity_date} {self.coupon} {self.par} {self.price} {self.strike}"
        if self.rating:
            out += f" {self.rating}"
        return out


class ExchangeableBonds(BaseModel):
    type: Literal["exchangeable_bonds"] = Field(
        default="exchangeable_bonds", frozen=True
    )
    issuer_ticker: str = fields.ticker_xbonds_issuer
    maturity_date: str = fields.maturity
    coupon: str = fields.coupon_bonds
    price: str = fields.price_bonds
    strike: str = fields.strike
    target_ticker: str = fields.ticker_xbonds_target
    rating: Optional[str] = fields.rating

    def __str__(self) -> str:
        out = f"EXCH:{self.issuer_ticker} {self.maturity_date} {self.coupon} {self.price} {self.strike} {self.target_ticker}"
        if self.rating:
            out += f" {self.rating}"
        return out


class GovernmentBonds(BaseModel):
    type: Literal["government_bonds"] = Field(default="government_bonds", frozen=True)
    country: str = fields.country_bonds
    maturity_date: str = fields.maturity
    coupon: str = fields.coupon_bonds
    price: str = fields.price_bonds
    rating: Optional[str] = fields.rating

    def __str__(self) -> str:
        out = f"GOV:{self.country} {self.maturity_date} {self.coupon} {self.price}"
        if self.rating:
            out += f" {self.rating}"
        return out


class InterestRateSwaps(BaseModel):
    type: Literal["interest_rate_swaps"] = Field(
        default="interest_rate_swaps", frozen=True
    )
    country: str = fields.country_irs
    maturity_date: str = fields.maturity
    payer_or_receiver: Literal["P", "R"] = fields.payer_or_receiver_irs
    spread: str = fields.spread

    def __str__(self) -> str:
        return f"SWAP:{self.country} {self.maturity_date} {self.payer_or_receiver}{self.spread}"


class CreditDefaultSwaps(BaseModel):
    type: Literal["credit_default_swaps"] = Field(
        default="credit_default_swaps", frozen=True
    )
    ticker: str = fields.ticker_cds
    maturity_date: str = fields.maturity
    payer_or_receiver: Literal["P", "R"] = fields.payer_or_receiver_cds
    spread: str = fields.spread

    def __str__(self) -> str:
        return f"CDS:{self.ticker} {self.maturity_date} {self.payer_or_receiver}{self.spread}"


class TermLoans(BaseModel):
    type: Literal["term_loans"] = Field(default="term_loans", frozen=True)
    ticker: str = fields.ticker_term_loans
    maturity_date: str = fields.maturity
    coupon: str = fields.coupon_term_loans
    price: str = fields.price_term_loans

    def __str__(self) -> str:
        return f"LOAN:{self.ticker} {self.maturity_date} {self.coupon} {self.price}"


class Warrants(BaseModel):
    type: Literal["warrants"] = Field(default="warrants", frozen=True)
    ticker: str = fields.ticker_warrants
    expiration_date: str = fields.expiration_warrants
    ratio: str = fields.ratio
    premium: str = fields.premium
    strike: str = fields.strike_warrants

    def __str__(self) -> str:
        return f"WRNT:{self.ticker} {self.expiration_date} {self.ratio} {self.premium} {self.strike}"


class Margin(BaseModel):
    type: Literal["margin"] = Field(default="margin", frozen=True)
    currency: str = fields.currency

    def __str__(self) -> str:
        return f"MARGIN:{self.currency}"


class Cash(BaseModel):
    type: Literal["cash"] = Field(default="cash", frozen=True)
    currency: str = fields.currency

    def __str__(self) -> str:
        return f"CASH:{self.currency}"


class Indices(BaseModel):
    type: Literal["indices"] = Field(default="indices", frozen=True)
    ticker: str = fields.ticker_indices

    def __str__(self) -> str:
        return f"IND:{self.ticker}"


class IndexOptions(BaseModel):
    type: Literal["index_options"] = Field(default="index_options", frozen=True)
    ticker: str = fields.ticker_index_options
    expiration_date: str = fields.expiration_index_options
    call_or_put: Literal["C", "P"] = fields.call_or_put
    strike: str = fields.strike_options
    premium: str = fields.premium

    def __str__(self) -> str:
        return f"IND:{self.ticker} {self.expiration_date} {self.call_or_put}{self.strike} {self.premium}"


class Repurchase(BaseModel):
    type: Literal["repurchase"] = Field(default="repurchase", frozen=True)
    instrument: Annotated[BrazilianFixedIncome, Field(discriminator="type")]

    def __str__(self) -> str:
        return self.instrument.repo_agreement


class Provision(BaseModel):
    type: Literal["provision"] = Field(default="provision", frozen=True)
    category: Literal["RECEIVABLES", "PAYABLES", "REDEMPTIONS"] = fields.category

    def __str__(self) -> str:
        return f"PROV:{self.category}"


class Symbol(BaseModel):
    """
    Wrapper model representing a financial instrument in the Everysk system.

    The `Symbol` class acts as a discriminated union over various financial instrument types.
    Each instrument is uniquely identified by its `type` field, which serves as the discriminator
    to allow Pydantic to correctly parse and validate the appropriate subclass instance.

    Attributes:
        instrument (Union[...]): A union of supported financial instrument models,
            such as BrazilianFutures, CorporateBonds, FXOptions, Cash, etc.
            The appropriate model is selected based on the 'type' field in the input data.

    Example:
        symbol = Symbol(
            instrument={
                'type': 'brazilian_futures',
                'ticker': 'IND',
                'expiration_date': '20251231',
                'price': '120000'
            }
        )
        print(str(symbol))  # -> 'BRFUT:IND 20251231 120000'
    """

    instrument: Annotated[
        Union[
            BrazilianFutures,
            BrazilianFixedIncome,
            USEquitiesETFsMutualFunds,
            ForeignEquities,
            EquityOptions,
            FXForward,
            FXOptions,
            Futures,
            FutureOptions,
            CorporateBonds,
            MunicipalBonds,
            ConvertibleBonds,
            ConvertiblePreferredBonds,
            ExchangeableBonds,
            GovernmentBonds,
            InterestRateSwaps,
            CreditDefaultSwaps,
            TermLoans,
            Warrants,
            Margin,
            Cash,
            Indices,
            IndexOptions,
            Repurchase,
            Provision,
        ],
        Field(discriminator="type"),
    ]

    def __str__(self):
        return str(self.instrument)


class Security(BaseModel):
    """
    Represents an Everysk security with its associated properties. It is used to validate
    the existence and validity of all required fields for a security in the Everysk system.

    Example usage:

    security = Security(
        id='sec_001',
        symbol=Symbol(
            instrument=BrazilianFutures(
                ticker='IND',
                expiration_date='20251231',
                price='120000'
            )
        ),
        quantity=10.0
    )

    >>> print(future.model_dump())
    >>> {'id': 'sec_001', 'symbol': 'BRFUT:IND 20251231 120000', 'quantity': 10.0}
    """

    id: str = fields.id_
    symbol: Symbol = fields.symbol
    quantity: float = fields.quantity
    isin: Optional[str] = fields.isin
    fx_rate: Optional[float] = fields.fx_rate
    market_price: float = fields.market_price
    market_value: float = fields.market_value

    def model_dump(self, **kwargs):
        """Override the model_dump method to include the symbol as a string in the output."""
        base = super().model_dump(**kwargs)
        base["symbol"] = str(self.symbol)
        return base

    class Config:
        """
        In Pydantic, the Config inner class is used to configure the behavior of the BaseModel.

        By default, Pydantic only allows standard Python types (eg., int, str, float, etc.) and
        other Pydantic models as field types. To use custom classes (like Symbol), arbitrary types
        must be explicitly allowed.
        """

        arbitrary_types_allowed = True



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import List, Literal, Optional

from mcp.server.fastmcp.prompts import base

from everysk_mcp.fields import definitions as fields


###############################################################################
# Implementation
###############################################################################
def register_prompts(mcp):
    """Register all prompt templates with the given MCP instance."""

    @mcp.prompt(
        name="Set Context",
        description="Sets the context for the assistant to use in the current session.",
    )
    def set_context(
        workspace: Optional[str] = fields.workspace,
    ) -> List[base.Message]:
        workspace = workspace or "main"

        return [
            base.AssistantMessage(
                f"I will be working on the workspace: `{workspace}`. I will use the workflows, report templates, reports, portfolios, datastores and files in this workspace to perform tasks and answer your questions."
            ),
            base.AssistantMessage(
                "This is a very workflow-centric assistant, so I will always try to use an existing workflow to complete tasks and to answer questions. To this end, my first action will always be to immediately list the workflows available in the workspace and keep this information ready to use."
            ),
            base.AssistantMessage(
                "Whenever possible, I will prefer to generate results on the screen. Only when asked I will produce HTML code for reports, tables and other deliverables."
            ),
            base.AssistantMessage(
                "Whenever possible, I will prefer to use symbols instead of sec_ids in the output to the user."
            ),
        ]

    @mcp.prompt(
        name="Set Pre-Trade Compliance",
        description="Sets the pre-trade compliance rules for the assistant to use in the current session.",
    )
    def set_pretrade_compliance(
        operation: Literal["buy", "sell"] = fields.operation,
        quantity: str = fields.quantity,
        security: str = fields.security_ticker,
    ) -> List[base.Message]:
        outflow = "outflow (ie, the sale of a security)"
        inflow = "inflow (ie, the purchase of a security)"

        return [
            base.AssistantMessage(
                f"I will now create a new portfolio, which will simulate a {operation} operation. Using single-entry bookkeeping, I will create this new portfolio simulating an {outflow if operation == 'sell' else inflow}."
            ),
            base.AssistantMessage(
                "After creating the new portfolio, I should check for the existence of a workflow with portfolio merging capabilities. I will use it to merge the new portfolio with the base portfolio. For efficiency, I should first retrieve the required parameter list for the workflow using the `workflow_prepare_run` tool."
            ),
            base.AssistantMessage(
                "Finally, I should check for the existence of a compliance workflow. I will use it to run the compliance metrics using the merged portfolio. For efficiency, I should first retrieve the required parameter list for the workflow using the `workflow_prepare_run` tool."
            ),
            base.AssistantMessage(
                "With everything set up, I will now proceed with the trade simulation."
            ),
            base.UserMessage(
                f"Operation: {operation}, security: {security}, quantity: {quantity}"
            ),
        ]



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from pydantic import Field

###############################################################################
# Implementation
###############################################################################

### Parameters ################################################################
id_ = Field(description="A unique identifier (UID) for a security.")

symbol = Field(description="Security symbol, according to Everysk's custom symbology")

quantity = Field(
    description="Numbers of shares or contracts. Positive quantities are longs and negative are shorts. For fixed income securities, mutual funds, swaps, cash and margin the amounts are defined in the security's currency.",
)

isin = Field(
    default=None,
    description="International Securities Identification Number (ISIN) for the security. This is an optional field and may not be available for all securities.",
    examples=[
        {"value": "US0378331005", "description": "ISIN for Apple Inc."},
        {"value": "US5949181045", "description": "ISIN for Microsoft Corporation"},
        {"value": None, "description": "No ISIN available for this security"},
    ],
)

fx_rate = Field(
    default=None,
    description="The price of the domestic currency stated in terms of another currency.",
)

market_price = Field(description="Current market price in local currency.")

market_value = Field(
    description="Current market value in local currency, calculated as `quantity * market_price`.",
)

### Asset code ################################################################
asset_code = Field(
    description="Ticker of brazilian asset code",
    examples=[
        {"value": "NTNB_20251231", "description": "Ticker for a NTN-B"},
        {"value": "NTNC_20251231", "description": "Ticker for a NTN-C"},
        {"value": "NTNF_20251231", "description": "Ticker for a NTN-F"},
        {"value": "LTN_20251231", "description": "Ticker for a LTN"},
        {"value": "LFT_20251231", "description": "Ticker for a LFT"},
        {"value": "BSA325", "description": "Ticker for a private debenture"},
        {"value": "CRA020001E5", "description": "Ticker for a CRA"},
        {"value": "LFSC1900081", "description": "Ticker for a LF"},
        {"value": "CDB017OBJWK", "description": "Ticker for a CDB"},
    ],
)

### Call or Put ###############################################################
call_or_put = Field(
    description="Indicates whether the option is a call or put option.",
    examples=[
        {"value": "C", "description": "This is a call option"},
        {"value": "P", "description": "This is a put option"},
    ],
)

### Category ##################################################################
category = Field(
    description="Type of provision, which can be RECEIVABLES, PAYABLES or REDEMPTIONS.",
    examples=[
        {"value": "RECEIVABLES", "description": "This is a receivables provision"},
        {"value": "PAYABLES", "description": "This is a payables provision"},
        {"value": "REDEMPTIONS", "description": "This is a redemptions provision"},
    ],
)

### Contract ##################################################################
contract = Field(
    description="Label for underlying future contract ",
    examples=[
        {"value": "CL", "description": "Crude Oil Futures"},
        {"value": "NG", "description": "Natural Gas Futures"},
    ],
)

### Country ###################################################################
country_bonds = Field(
    description='ISO code for the country of the government bond. For example, "US" for United States, "BR" for Brazil.',
    examples=[
        {
            "value": "US",
            "description": "US is the ISO code for United States government bonds",
        },
        {
            "value": "BR",
            "description": "BR is the ISO code for Brazilian government bonds",
        },
        {
            "value": "DE",
            "description": "DE is the ISO code for German government bonds",
        },
        {
            "value": "JP",
            "description": "JP is the ISO code for Japanese government bonds",
        },
    ],
)

country_irs = Field(
    description="ISO Code for the issuer government (US only).",
    examples=[
        {
            "value": "US",
            "description": "US is the ISO code for United States interest rate swaps",
        }
    ],
)

### Coupon ####################################################################
coupon_bonds = Field(
    description="The coupon rate of the bond, expressed as a percentage.",
    examples=[
        {"value": "5.25", "description": "Coupon rate of the bond is 5.25%"},
        {"value": "10.00", "description": "Coupon rate of the bond is 10.00%"},
    ],
)

coupon_term_loans = Field(
    description="The spread over Libor for the loan, expressed as a percent amount.",
    examples=[
        {
            "value": "5.25",
            "description": "Spread over Libor for the term loan is 5.25%",
        },
        {
            "value": "10.00",
            "description": "Spread over Libor for the term loan is 10.00%",
        },
    ],
)

### Cross currency ############################################################
cc_fx_forward = Field(
    description="ISO code for the cross currency pair. Users can define any combination.",
    examples=[
        {"value": "MXNUSD", "description": "Mexican Peso to US Dollar"},
        {"value": "EURAUD", "description": "Euro to Australian Dollar"},
    ],
)

cc_fx_options = Field(
    description="ISO code for the cross currency pair. Users can define any combination. The option symbology is always in reference to the first currency, so a put on MXNUSD is a put on Mexican peso, call on US dollar.",
    examples=[
        {"value": "MXNUSD", "description": "Mexican Peso to US Dollar"},
        {"value": "EURAUD", "description": "Euro to Australian Dollar"},
    ],
)

### Currency ##################################################################
currency = Field(
    description="ISO code for the currency of the cash account.",
    examples=[
        {"value": "USD", "description": "US Dollar"},
        {"value": "EUR", "description": "Euro"},
        {"value": "BRL", "description": "Brazilian Real"},
    ],
)

### Expiration date ###########################################################
expiration_contract = Field(
    description="String in YYYYMMDD format indicating the expiration date for the contract.",
    examples=[
        {
            "value": "20250101",
            "description": "Expiration date for the contract is January 1, 2025",
        },
        {
            "value": "20251231",
            "description": "Expiration date for the contract is December 31, 2025",
        },
    ],
)

expiration_index_options = Field(
    description="String in YYYYMMDD format indicating the expiration date for the option.",
    examples=[
        {
            "value": "20250101",
            "description": "Expiration date for the index option is January 1, 2025",
        },
        {
            "value": "20251231",
            "description": "Expiration date for the index option is December 31, 2025",
        },
    ],
)

expiration_warrants = Field(
    description="String in YYYYMMDD format indicating the expiration date for the warrant.",
    examples=[
        {
            "value": "20250101",
            "description": "Expiration date for the warrant is January 1, 2025",
        },
        {
            "value": "20251231",
            "description": "Expiration date for the warrant is December 31, 2025",
        },
    ],
)

### Forward rate ##############################################################
forward_rate = Field(
    description="Indicates the contracted exchange rate.",
    examples=[
        {
            "value": "0.045",
            "description": "This quantity is always in respect with the first currency in the pair. Thus, for MXNUSD, 0.045 with a quantity of -40e6 means that 40 milllion of Mexican pesos are to be exchanged by 1.8 million US dollars on the specified date.",
        }
    ],
)

### Issuer ####################################################################
issuer_muni_bonds = Field(
    description="Ticker for issuer of bond (US only).",
    examples=[
        {"value": "PA", "description": "Ticker for Pennsylvania"},
    ],
)

### Maturity date #############################################################
maturity = Field(
    description="String in YYYYMMDD format indicating the maturity date for the security.",
    examples=[
        {
            "value": "20250101",
            "description": "Maturity date for the security is January 1, 2025",
        },
        {
            "value": "20251231",
            "description": "Maturity date for the security is December 31, 2025",
        },
    ],
)

### MIC #######################################################################
mic = Field(
    description="Market Identifier Code (MIC) is the ISO code for the global exchange.",
    examples=[
        {"value": "XETR", "description": "MIC for Frankfurt Stock Exchange"},
        {"value": "XTKS", "description": "MIC for Tokyo Stock Exchange"},
        {"value": "BVMF", "description": "MIC for B3 (Brazilian Stock Exchange)"},
    ],
)

### Par #######################################################################
par = Field(
    description="The par value of the convertible preferred bond, which is the face value of the bond.",
    examples=[
        {
            "value": "100",
            "description": "Par value of the convertible preferred bond is $100.00",
        },
        {
            "value": "105.5",
            "description": "Par value of the convertible preferred bond is $105.50",
        },
    ],
)

### Payer or Receiver #########################################################
payer_or_receiver_irs = Field(
    description="P indicates a payer swap (pay fixed/receive floating) /  R indicates a receiver swap (receive fixed/pay floating)",
    examples=[
        {"value": "P", "description": "This is a payer swap"},
        {"value": "R", "description": "This is a receiver swap"},
    ],
)

payer_or_receiver_cds = Field(
    description="P indicates a payer CDS (pay premium/receive protection) /  R indicates a receiver CDS (receive premium/pay protection)",
    examples=[
        {"value": "P", "description": "This is a payer CDS"},
        {"value": "R", "description": "This is a receiver CDS"},
    ],
)

### Premium ###################################################################
premium = Field(
    description="The premium of the contract, which is the price paid for it.",
    examples=[
        {"value": "5.25", "description": "Premium of the contract is $5.25"},
        {"value": "10.00", "description": "Premium of the contract is $10.00"},
    ],
)

### Price #####################################################################
price = Field(
    description="The asset price.",
    examples=[
        {"value": "5.25", "description": "Unit price of the asset is $5.25"},
        {"value": "10.00", "description": "Unit price of the asset is $10.00"},
    ],
)

price_bonds = Field(
    description="The price of the bond, provided as an absolute price, not percent of par.",
    examples=[
        {"value": "100", "description": "Price of the bond is $100.00"},
        {"value": "105.5", "description": "Price of the bond is $105.50"},
    ],
)

price_term_loans = Field(
    description="The price of the term loan, provided as an absolute price, not percent of par.",
    examples=[
        {"value": "100", "description": "Price of the term loan is $100.00"},
        {"value": "105.5", "description": "Price of the term loan is $105.50"},
    ],
)

### Rating ####################################################################
rating = Field(
    default=None,
    description="The bond rating formatted in S&P rating syntax e.g., AA, BBB+, etc. Used when the issuer curve is not available or multiple ratings exist for an issuer.",
    examples=[
        {"value": "AA", "description": "Credit rating of the convertible bond is AA"},
        {"value": None, "description": "No credit rating available for this bond"},
    ],
)

### Ratio #####################################################################
ratio = Field(
    description="Indicates the number of common shares per warrant upon conversion.",
    examples=[
        {
            "value": "1",
            "description": "One warrant allows the purchase of one share of the underlying stock",
        },
        {
            "value": "0.5",
            "description": "One warrant allows the purchase of half a share of the underlying stock",
        },
    ],
)

### Spread ####################################################################
spread = Field(
    description="The spread of the swap, expressed as a percentage.",
    examples=[
        {"value": "0.25", "description": "Spread of the swap is 0.25%"},
        {"value": "0.50", "description": "Spread of the swap is 0.50%"},
    ],
)

### Strike ####################################################################
strike = Field(
    description="The strike price of the bond, which is the price at which it can be converted.",
    examples=[
        {"value": "150", "description": "Strike price of the bond is $150.00"},
        {"value": "120", "description": "Strike price of the bond is $120.00"},
    ],
)

strike_options = Field(
    description="The strike of the option contract, which is the price at which the option can be exercised.",
    examples=[
        {
            "value": "150",
            "description": "Strike price of the option contract is $150.00",
        },
        {
            "value": "120",
            "description": "Strike price of the option contract is $120.00",
        },
    ],
)

strike_warrants = Field(
    description="The strike price of the warrant, which is the price at which the warrant can be exercised.",
    examples=[
        {"value": "150", "description": "Strike price of the warrant is $150.00"},
        {"value": "120", "description": "Strike price of the warrant is $120.00"},
    ],
)

### Ticker ####################################################################
ticker_br = Field(
    description="Ticker of brazilian asset code",
    examples=[
        {
            "value": "AFS",
            "description": "AFS is the ticker for the South African Rand Future",
        },
        {
            "value": "DAP",
            "description": "DAP is the ticker for the ID x IPCA coupon Future",
        },
        {"value": "DDI", "description": "DDI is the ticker for the Selic Rate Future"},
        {"value": "DI1", "description": "DI1 is the ticker for the DI Rate Future"},
        {
            "value": "IND",
            "description": "IND is the ticker for the Ibovespa Index Future",
        },
        {
            "value": "ISP",
            "description": "ISP is the ticker for the S&P500 Index Future",
        },
        {
            "value": "OC1",
            "description": "OC1 is the ticker for the IDI x U.S. Dollar Spread Future",
        },
        {
            "value": "WIN",
            "description": "WIN is the ticker for the Mini Ibovespa Index Future",
        },
        {
            "value": "WSP",
            "description": "WSP is the ticker for the Mini S&P500 Index Future",
        },
        {
            "value": "CNH",
            "description": "CNH is the ticker for the Chinese Iuan (Type A) Future",
        },
        {"value": "DOL", "description": "DOL is the ticker for the US Dollar Future"},
        {
            "value": "GBR",
            "description": "GBR is the ticker for the Sterling Pound Future",
        },
        {
            "value": "EUP",
            "description": "EUP is the ticker for the Euro (Type B) Future",
        },
        {"value": "EUR", "description": "EUR is the ticker for the Euro Future"},
        {"value": "IAP", "description": "IAP is the ticker for the IPCA Spread Future"},
        {
            "value": "JAP",
            "description": "JAP is the ticker for the Japanese Yen Future",
        },
        {
            "value": "WDO",
            "description": "WDO is the ticker for the Mini US Dollar Future",
        },
        {
            "value": "AUS",
            "description": "AUS is the ticker for the Australian Dollar Future",
        },
        {
            "value": "CAN",
            "description": "CAN is the ticker for the Canadian Dollar Future",
        },
        {
            "value": "CHL",
            "description": "CHL is the ticker for the Chilean Peso Future",
        },
        {
            "value": "CNY",
            "description": "CNY is the ticker for the Chinese Iuan Future",
        },
        {
            "value": "MEX",
            "description": "MEX is the ticker for the Mexican Peso Future",
        },
        {
            "value": "NOK",
            "description": "NOK is the ticker for the Norwegian krone Future",
        },
        {
            "value": "RUB",
            "description": "RUB is the ticker for the Russian Ruble Future",
        },
        {"value": "SWI", "description": "SWI is the ticker for the Swiss Franc Future"},
        {
            "value": "TUQ",
            "description": "TUQ is the ticker for the Turkish Pound Future",
        },
    ],
)

ticker_bonds = Field(
    description="Ticker for issuer of bond (US only).",
    examples=[
        {"value": "NFLX", "description": "Ticker for Netflix, Inc."},
        {"value": "COMM", "description": "Ticker for Comcast Corporation"},
    ],
)

ticker_cds = Field(
    description="Ticker for the company credit protection is being bought/sold (US only).",
    examples=[
        {"value": "NFLX", "description": "Ticker for Netflix, Inc."},
        {"value": "COMM", "description": "Ticker for Comcast Corporation"},
    ],
)

ticker_equity_options = Field(
    description=" Simple ticker for US exchanges or ticker and exchange (MIC) for global equities.",
    examples=[
        {
            "value": "IBM",
            "description": "Ticker for International Business Machines Corporation",
        },
        {
            "value": "PETR3:BVMF",
            "description": "Ticker and MIC for Petrobras S.A. on B3",
        },
    ],
)

ticker_xbonds_issuer = Field(
    description="Ticker for issuer of exchangeable bond (US only).",
    examples=[
        {"value": "NFLX", "description": "Ticker for Netflix, Inc."},
        {"value": "COMM", "description": "Ticker for Comcast Corporation"},
    ],
)

ticker_xbonds_target = Field(
    description="Ticker for the target company of the exchangeable bond (the company the bond can be converted into).",
    examples=[
        {
            "value": "AAPL",
            "description": "Ticker for Apple Inc., the target company of the exchangeable bond",
        },
        {
            "value": "GOOGL",
            "description": "Ticker for Alphabet Inc., the target company of the exchangeable bond",
        },
    ],
)

ticker_foreign_equities = Field(
    description="Ticker of foreign asset code.",
    examples=[
        {"value": "SIE", "description": "Ticker for Siemens AG on XETRA"},
        {
            "value": "6758",
            "description": "Ticker for Sony Group Corporation on Tokyo Stock Exchange",
        },
        {"value": "PETR3", "description": "Ticker for Petrobras S.A. on B3"},
    ],
)

ticker_index_options = Field(
    description="Ticker for the index option.",
    examples=[
        {"value": "SPX", "description": "S&P 500 Index Option"},
        {"value": "NDX", "description": "NASDAQ-100 Index Option"},
        {"value": "DJI", "description": "Dow Jones Industrial Average Option"},
    ],
)

ticker_indices = Field(
    description="Ticker for the index.",
    examples=[
        {"value": "SPX", "description": "S&P 500 Index"},
        {"value": "NDX", "description": "NASDAQ-100 Index"},
        {"value": "DJI", "description": "Dow Jones Industrial Average"},
    ],
)

ticker_term_loans = Field(
    description="Ticker for the company that issued the term loan (US only).",
    examples=[
        {"value": "NFLX", "description": "Ticker for Netflix, Inc."},
        {"value": "COMM", "description": "Ticker for Comcast Corporation"},
    ],
)

ticker_us = Field(
    description="Ticker of US asset code. For securities listed in the US exchanges (XNAS, XNYS, ARCX, XASE) the MIC - Market Identification Code - is not necessary.",
    examples=[
        {"value": "AAPL", "description": "Ticker for Apple Inc."},
        {
            "value": "IBM",
            "description": "Ticker for International Business Machines Corporation",
        },
        {"value": "USO", "description": "Ticker for United States Oil Fund"},
        {
            "value": "BAGIX",
            "description": "Ticker for Baird Aggregate Bond Index Fund Institutional Class",
        },
    ],
)

ticker_warrants = Field(
    description="Simple Ticker for US exchanges.",
    examples=[
        {"value": "EXXIW", "description": "Ticker for the warrant on EXXI"},
    ],
)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Optional

from everysk.api import Report
from everysk.api.utils import EveryskList
from mcp.server.fastmcp import FastMCP

from everysk_mcp.fields import definitions as fields


###############################################################################
# Implementation
###############################################################################
def register_report_tools(mcp: FastMCP) -> None:
    """Register all report tools with the given MCP instance."""

    @mcp.tool(
        description="Returns a list of reports previously generated. The reports are returned in sorted order, with the most recent report appearing first."
    )
    def report_list(
        query: Optional[str] = fields.query,
        workspace: Optional[str] = fields.workspace,
        page_size: Optional[int] = fields.page_size,
    ) -> EveryskList[Report]:
        result = Report.list(query=query, workspace=workspace, page_size=page_size)
        return result

    @mcp.tool(
        description="Retrieves the details of an existing report by supplying the report's id."
    )
    def report_retrieve(
        id: str = fields.id_, workspace: Optional[str] = fields.workspace
    ) -> Report:
        result = Report.retrieve(id=id, workspace=workspace)
        return result



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import List, Dict

from mcp.server.fastmcp import FastMCP


###############################################################################
# Implementation
###############################################################################
def register_static_resources(mcp: FastMCP) -> None:
    """ Register all static resources with the given MCP instance. """

    @mcp.resource(
        uri='config://color-palette', 
        name='Get color palette', 
        description='Returns the color palette used in the application.'
    )
    def get_color_palette() -> Dict[str, List[str]]:
        return {
            'primary': ['#FFFFFF', '#121212', '#FF832A'],
            'secondary': ['#FFB884', '#232323'],
            'tertiary': ['#F4F4F4', '#7B7B7B'],
        }




###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from pydantic import Field

###############################################################################
# Implementation
###############################################################################
aggregation = Field(
    default="position",
    description="Aggregation computes individual security exposures and aggregates them according to the supported criteria",
    examples=[
        {"value": "country", "description": "Aggregate by country"},
        {"value": "currency", "description": "Aggregate by currency"},
        {"value": "dividend_yield", "description": "Aggregate by dividend yield"},
        {"value": "duration", "description": "Aggregate by duration"},
        {"value": "exposure", "description": "Aggregate by exposure"},
        {"value": "fixed_income", "description": "Aggregate fixed income securities"},
        {"value": "gics_sector", "description": "Aggregate by GICS sector"},
        {"value": "implied_rating", "description": "Aggregate by implied rating"},
        {"value": "liquidity", "description": "Aggregate by liquidity"},
        {
            "value": "market_capitalization",
            "description": "Aggregate by market capitalization",
        },
        {"value": "position", "description": "Aggregate by position"},
        {"value": "security_type", "description": "Aggregate by security type"},
        {
            "value": "security_type_refined",
            "description": "Refined security type aggregation",
        },
        {"value": "sector", "description": "Aggregate by sector"},
        {"value": "total_esg", "description": "Aggregate total ESG scores"},
    ],
)

base_currency = Field(
    default="USD",
    description="3-letter ISO 4217 code for currency.",
    examples=[
        {"value": "USD", "description": "Base currency of the entity is US Dollar"},
        {"value": "EUR", "description": "Base currency of the entity is Euro"},
        {
            "value": "BRL",
            "description": "Base currency of the entity is Brazilian Real",
        },
    ],
)

compute_notional = Field(
    default=False,
    description="Determines whether sensitivities should be weighted by the notional exposure or unitized. True:Weight sensitivities by notional exposure. False: Unitize sensitivities.",
    examples=[
        {
            "value": True,
            "description": "Compute sensitivities weighted by notional exposure",
        },
        {"value": False, "description": "Compute unitized sensitivities"},
    ],
)

confidence = Field(
    default="95%",
    description="This parameter determines the confidence level for calculating CVaR. Values accepted: 1sigma, 2sigma, 3sigma, 85%, 90%, 95%, 97%, 99%",
    examples=[
        {"value": "90%", "description": "Confidence level of 90%"},
        {"value": "1sigma", "description": "1-sigma confidence level"},
    ],
)

content_type = Field(
    default="text/plain",
    description="A string to identify the file's format.",
    examples=[
        {"value": "text/plain", "description": "Plain text file"},
        {"value": "application/pdf", "description": "PDF file"},
        {"value": "image/png", "description": "PNG image file"},
        {"value": "application/json", "description": "JSON file"},
    ],
)

correlation_half_life = Field(
    default=6,
    description="Half life of correlation information in months: 0 (no decay), 2, 6, 12, 24 or 48.",
    examples=[
        {"value": 0, "description": "No correlation half-life applied"},
        {"value": 2, "description": "Correlation half-life of 2 days"},
        {"value": 6, "description": "Correlation half-life of 6 days"},
    ],
)

data_file = Field(
    description="The content that will be stored inside the file. It must be a text string.",
    examples=[
        {
            "value": "Hello, World!",
            "description": 'A simple text file with the content "Hello, World!"',
        },
    ],
)

data_datastore = Field(
    description="A list of lists representing the data to be stored in the datastore. The first inner list represents the column names, and the subsequent inner lists represent the rows of data. Each inner list should contain values that are either strings, integers, floats, or booleans.",
    examples=[
        {
            "value": [
                ["date", "ticker", "price"],
                ["2023-01-01", "AAPL", 150.0],
                ["2023-01-02", "AAPL", 152.5],
            ],
            "description": "A simple datastore with two columns: date and ticker, and two rows of data.",
        },
        {
            "value": [
                ["name", "age", "is_active"],
                ["Alice", 30, True],
                ["Bob", 25, False],
            ],
            "description": "A datastore with three columns: name, age, and is_active, and two rows of data.",
        },
    ],
)

date = Field(
    default=None,
    description="A string in YYYYMMDD format indicating the date of the data being stored in the entity. If not provided, the current date will be used.",
    examples=[
        {
            "value": "20250101",
            "description": "Data date for the entity is January 1, 2025",
        },
        {
            "value": "20251231",
            "description": "Data date for the entity is December 31, 2025",
        },
        {
            "value": None,
            "description": "No specific date provided, current date will be used",
        },
    ],
)

date_portfolio = Field(
    default=None,
    description="Portfolio date in the format: YYYYMMDD. The Portfolio Date is important because it instructs the API to use the market conditions and security prices prevailing on that date. Therefore, historical portfolios will be calculated without look ahead. In order to run the portfolio using market conditions prevailing today, use null.",
    examples=[
        {
            "value": "20250101",
            "description": "Data date for the portfolio is January 1, 2025",
        },
        {
            "value": "20251231",
            "description": "Data date for the portfolio is December 31, 2025",
        },
        {
            "value": None,
            "description": "No specific date provided, current date will be used",
        },
    ],
)

description = Field(
    description="Provides detailed information about an entity.",
    examples=[
        {
            "value": "This is a sample text file.",
            "description": "A simple text file with a description",
        },
        {
            "value": "Annual financial report for 2023",
            "description": "A PDF report with a detailed description",
        },
        {
            "value": "This datastore contains financial data for analysis.",
            "description": "A simple description for the datastore",
        },
        {
            "value": "Datastore for storing historical stock prices and trading volumes",
            "description": "A detailed description for the datastore",
        },
        {
            "value": "This portfolio contains financial data for analysis.",
            "description": "A simple description for the portfolio",
        },
        {
            "value": "Portfolio for storing historical stock prices and trading volumes",
            "description": "A detailed description for the portfolio",
        },
    ],
)

end_date = Field(
    default=None,
    description="The end date for filtering entities, in YYYYMMDD format.",
    examples=[
        {
            "value": "20251231",
            "description": "Filter entities created on or before December 31, 2025",
        },
        {"value": None, "description": "No end date filtering applied"},
    ],
)

ewma = Field(
    default=False,
    description="If True, the exponentially weighted moving average will be applied to volatilities.",
    examples=[
        {"value": True, "description": "Use exponentially weighted moving average"},
        {"value": False, "description": "Use simple average"},
    ],
)

exponential_decay = Field(
    default=0.94,
    description='Weighting factor determining the rate at which "older" data enter into the calculation of the EWMA. It accepts float values from 0 to 1. This parameter is only used when EWMA is True.',
    examples=[
        {"value": 0.94, "description": "Exponential decay factor of 0.94"},
        {"value": 0.90, "description": "Exponential decay factor of 0.90"},
    ],
)

filter_ = Field(
    default=None,
    description="Filters are powerful constructs to return the unique identifiers of certain positions in the portfolio, satisfying a specific criteria. When used, they isolate the behavior of a unique group, compared to the whole. For example: how illiquid securities might react to an oil shock. When present it runs a pre-processing calculation to select identifiers that satisfy certain criteria. It can be used to highlight the properties of a subset of the portfolio, for example: only fixed income securities. Calculations are performed for the whole portfolio and only displayed for the securities that satisfy the filter.",
    examples=[
        {
            "value": "exposure('long') and currency('foreign securities')",
            "description": "Example using logical operator 'and'. Retrieve identifiers for positions that are long and also denominated in a currency different than the base currency of the portfolio",
        },
        {
            "value": "exposure('short') or liquidity('>300%')",
            "description": "Example using logical operator 'or'. Retrieve any shorts or illiquid positions",
        },
        {
            "value": "not(type('equity option long delta'))",
            "description": "Example using logical operator 'not'. Retrieve all positions that are not long options",
        },
        {
            "value": "currency('foreign securities') and (not(exposure('short') or liquidity('>300%')))",
            "description": "Example using logical operators combined",
        },
        {"value": None, "description": "Use the whole portfolio"},
    ],
)

historical_days = Field(
    default=252,
    description="Number of business days used to calculate the covariance for the primitive risk factors: 0 (no decay), 2, 6, 12, 24 or 48.",
    examples=[
        {
            "value": 252,
            "description": "Calculate risk attribution over the last 252 trading days",
        },
        {
            "value": 30,
            "description": "Calculate risk attribution over the last 30 trading days",
        },
    ],
)

horizon = Field(
    default=5,
    description="Simulates the behavior of each security via their underlying risk factors. It accepts: 1, 5, 20 or 60",
    examples=[
        {"value": 1, "description": "Simulate risk attribution for 1 day"},
        {"value": 5, "description": "Simulate risk attribution for 5 days"},
    ],
)

id_ = Field(
    description="A unique identifier (UID) for an entity.",
    examples=[
        {
            "value": "dats_52KiJy2yFvosTtWM5aUJcOs1R",
            "description": "A datastore's id will always look like this.",
        },
        {
            "value": "file_52KiJy2yFvosTtWM5aUJcOs1R",
            "description": "A file's id will always look like this.",
        },
        {
            "value": "port_52KiJy2yFvosTtWM5aUJcOs1R",
            "description": "A portfolio's id will always look like this.",
        },
        {
            "value": "repo_52KiJy2yFvosTtWM5aUJcOs1R",
            "description": "A report's id will always look like this.",
        },
        {
            "value": "rtpl_52KiJy2yFvosTtWM5aUJcOs1R",
            "description": "A report template's id will always look like this.",
        },
        {
            "value": "wrkf_1hud1QQNR2z7m9IPLrSylpOMN",
            "description": "A workflow's id will always look like this.",
        },
        {
            "value": "wfex_N4buYFKsRDtlvavf3TvnnCkCh",
            "description": "An execution's id will always look like this.",
        },
    ],
)

id_exclusive = Field(
    default=None,
    description="It is the ID of and existing portfolio. When passing a portfolio's ID as parameter the calculation will be performed using the given portfolio's securities, date, base currency and nlv. The securities parameter is no longer required.",
    examples=[
        {
            "value": "port_52KiJy2yFvosTtWM5aUJcOs1R",
            "description": "A portfolio's id will always look like this.",
        }
    ],
)

limit = Field(
    default=100,
    description="The maximum number of entities to retrieve.",
    examples=[
        {"value": 100, "description": "Default limit, returns up to 100 entities"},
        {"value": 50, "description": "Returns up to 50 entities"},
    ],
)

link_uid = Field(
    default=None,
    description="A unique identifier (UID) for a link associated with the entity. If provided, only entities linked to this UID will be returned.",
    examples=[
        {
            "value": "link_1234567890abcdef",
            "description": "Filter by a specific link UID",
        },
        {"value": None, "description": "No link UID filtering applied"},
    ],
)

magnitude = Field(
    default="-0.10",
    description="The magnitude of the shock. The parameter magnitude is an instantaneous shock as a decimal number: -10% = -0.10. Certain rates/spreads can also be shocked and the magnitude of the shock is an absolute widening or tightening of the rate/spread. Example 1: Shock: IND: SPX | Magnitude: -1% | Current level: 2821| New level: 2792. Example 2: Shock: IND: LIBOR3M | Magnitude: -1% | Current level: 1.77| New level: 0.77. In these examples, the shock in SPX is a percent change, whereas the libor shock is an absolute shift.",
    examples=[
        {"value": "-0.05", "description": "Apply a 5 percent shock"},
        {"value": "-0.10", "description": "Apply a 10 percent shock"},
    ],
)

name = Field(
    description="A string to identify an entity besides its id. Feel free to use a meaningful name.",
    examples=[
        {"value": "my_file.txt", "description": "A simple text file named my_file.txt"},
        {"value": "report.pdf", "description": "A PDF report file named report.pdf"},
        {
            "value": "my_datastore",
            "description": "A simple datastore named my_datastore",
        },
        {
            "value": "financial_data_datastore",
            "description": "A datastore containing financial data",
        },
        {
            "value": "my_portfolio",
            "description": "A simple portfolio named my_portfolio",
        },
        {
            "value": "financial_data_portfolio",
            "description": "A portfolio containing financial data",
        },
    ],
)

ndays = Field(
    default=252,
    description="Number of historical business days used in calculation when weekly is False.",
    examples=[
        {
            "value": 252,
            "description": "Calculate tracking error over the last 252 trading days",
        },
        {
            "value": 30,
            "description": "Calculate tracking error over the last 30 trading days",
        },
    ],
)

nlv = Field(
    default=None,
    description="The net liquidating value of the entity (also called NAV). When not supplied by user, the API will calculate the net liquidating value of all the cash securities. Securities traded on margin are assumed to have NLV zero. Supplying an NLV effectively reconciles the margin.",
    examples=[
        {
            "value": 100000.0,
            "description": "Net Liquidation Value of the entity is $100,000.00",
        },
        {
            "value": None,
            "description": "No NLV provided, it will be calculated based on the securities",
        },
    ],
)

nweeks = Field(
    default=200,
    description="Number of historical weeks used in calculation when weekly is True.",
    examples=[
        {
            "value": 200,
            "description": "Calculate tracking error over the last 200 weeks",
        },
        {"value": 52, "description": "Calculate tracking error over the last 52 weeks"},
    ],
)

operation = Field(
    description='Type of operation to be performed, either "buy" or "sell".',
)

page_size = Field(
    default=100,
    description="Set the number of entities that will be listed per page.",
    examples=[
        {
            "value": 100,
            "description": "Default page size, returns 100 entities per page",
        },
        {"value": 50, "description": "Returns 50 entities per page"},
    ],
)

parameters = Field(
    default={},
    description="Object sent to be used inside the workflow that will be started through the API. Usually this object contains information and data required by the workers inside the workflow in order to execute properly. If the workers inside the target workflow do not require external inputs you can pass a empty object as parameter.",
    examples=[
        {
            "value": {"param1": "value1", "param2": 42},
            "description": "Parameters to be passed to the workflow",
        },
        {
            "value": {},
            "description": "No parameters provided, using default values if available",
        },
    ],
)

projection = Field(
    default=["IND:SPX"],
    description="User supplied array of securities to be used as a top-down factor model. Maximum number of elements in the projection array is 15.",
    examples=[
        {"value": ["IND:SPX"], "description": "Projection using the S&P 500 index"},
        {
            "value": ["IND:NDX", "IND:DJI"],
            "description": "Projection using multiple indices",
        },
    ],
)

quantity = Field(description="Quantity of the security to be traded")

query = Field(
    default=None,
    description="Request a list of entities filtering it by `name` or `tag`. When using a `tag` to perform a query, each term must include a hashtag prefix.",
    examples=[
        {"value": None, "description": "Return all entities"},
        {"value": "my-entity", "description": "Filter by entity name"},
        {
            "value": "#tag1 #tag2",
            "description": "Filter by tags, each tag must start with a hashtag.",
        },
    ],
)

risk_measure = Field(
    default="vol",
    description="Specifies the forward looking portfolio risk property being measured by MCTR. vol: forward looking annualized volatility of the P&L distribution for the portfolio. var: Value-at-Risk of the P&L distribution for the portfolio. cvar: Conditional Value-at-Risk of the P&L distribution for the portfolio.",
    examples=[
        {"value": "vol", "description": "Calculate volatility"},
        {"value": "var", "description": "Calculate value at risk"},
        {"value": "cvar", "description": "Calculate conditional value at risk"},
    ],
)

sampling = Field(
    default=1,
    description="Sampling represents the frequency that historical prices and rates (spreads) are collected in order to compute the invariant risk factors. In order to sample daily, use 1. In order to sample weekly (non-overlapping), pass 5. It accepts: 1 or 5.",
    examples=[
        {"value": 1, "description": "Sample daily"},
        {"value": 5, "description": "Sample weekly (non-overlapping)"},
    ],
)

security_ticker = Field(
    description="Ticker of the security to be traded",
)

security_list = Field(
    description="It is an array of securities to describe the securities in the portfolio. Each object represents a security with a unique id, symbol, quantity",
)

security_list_exclusive = Field(
    default=None,
    description="It is an array describing the securities in the portfolio. Not required if portfolio_id is being provided.",
)

sensitivity_type = Field(
    default="delta",
    description="The type of sensitivity to be calculated.",
    examples=[
        {"value": "average_life", "description": "Calculate average life sensitivity"},
        {
            "value": "cs01",
            "description": "Calculate CS01 (Credit Spread 01) sensitivity",
        },
        {"value": "delta", "description": "Calculate delta sensitivity"},
        {"value": "duration", "description": "Calculate duration sensitivity"},
        {
            "value": "dv01",
            "description": "Calculate DV01 (dollar value of 01) sensitivity",
        },
        {
            "value": "effective_duration",
            "description": "Calculate effective duration sensitivity",
        },
        {"value": "gamma", "description": "Calculate gamma sensitivity"},
        {
            "value": "macaulay_duration",
            "description": "Calculate Macaulay duration sensitivity",
        },
        {
            "value": "modified_duration",
            "description": "Calculate modified duration sensitivity",
        },
        {
            "value": "rolled_average_life",
            "description": "Calculate rolled average life sensitivity",
        },
        {"value": "theta", "description": "Calculate theta sensitivity"},
        {"value": "vega", "description": "Calculate vega sensitivity"},
        {"value": "volatility", "description": "Calculate volatility sensitivity"},
    ],
)

shock = Field(
    default="IND:SPX",
    description="The security being used for the stress test.",
    examples=[
        {"value": "IND:SPX", "description": "Apply a shock to the S&P 500 index"},
        {"value": "IND:NDX", "description": "Apply a shock to the NASDAQ-100 index"},
    ],
)

synchronous = Field(
    default=True,
    description="When True, the workflow will be run synchronously, meaning the API will wait for the workflow to finish before returning the result. When False, the workflow will be run asynchronously, and the API will return immediately with a reference to the workflow execution.",
    examples=[
        {
            "value": True,
            "description": "Run the workflow synchronously, waiting for it to finish",
        },
        {
            "value": False,
            "description": "Run the workflow asynchronously, returning immediately",
        },
    ],
)

start_date = Field(
    default=None,
    description="The start date for filtering entities, in YYYYMMDD format.",
    examples=[
        {
            "value": "20250101",
            "description": "Filter entities created on or after January 1, 2025",
        },
        {"value": None, "description": "No start date filtering applied"},
    ],
)

tags = Field(
    default=[],
    description="Sequence of hashtags used to label or filter the related entity. Any sequence of lowercase character, numbers and underscore might be used.",
    examples=[
        {"value": ["tag1", "tag2"], "description": "Filter by multiple tags"},
        {"value": [], "description": "No tag filtering applied"},
    ],
)

use_cashflow = Field(
    default=True,
    description="For fixed income securities, when True, maps each cashflow event (interest, principal amortization) to various key points in the respective interest rate curve. When flag is False, it uses the Macaulay duration to map bond price to neighboring key points in the respective interest rate curve.",
    examples=[
        {"value": True, "description": "Use cashflow data for risk attribution"},
        {
            "value": False,
            "description": "Use current market prices for risk attribution",
        },
    ],
)

volatility_half_life = Field(
    default=2,
    description="Half life of volatility information in months: 0 (no decay), 2, 6, 12, 24 or 48.",
    examples=[
        {"value": 0, "description": "No volatility half-life applied"},
        {"value": 2, "description": "Volatility half-life of 2 days"},
        {"value": 6, "description": "Volatility half-life of 6 days"},
    ],
)

weekly = Field(
    default=True,
    description="The Calculation API will use weekly return to calculate the tracking error.",
    examples=[
        {"value": True, "description": "Calculate weekly tracking error"},
        {"value": False, "description": "Calculate daily tracking error"},
    ],
)

with_data = Field(
    default=True,
    description="When True, the the data inside the entity will be returned in the api response.",
    examples=[
        {"value": True, "description": "Return the entity with its data included"},
        {"value": False, "description": "Return the entity without its data"},
    ],
)

with_securities = Field(
    default=True,
    description="When True, the securities inside the entity will be returned in the api response.",
    examples=[
        {
            "value": True,
            "description": "Return the entity with its securities included",
        },
        {"value": False, "description": "Return the entity without its securities"},
    ],
)

workspace = Field(
    default="main",
    description="Determines on which workspace the request will be made. Default is `main`.",
    examples=[
        {"value": "main", "description": "Default workspace"},
        {
            "value": "my-workspace",
            "description": "A custom workspace where the files are listed",
        },
    ],
)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import List, Literal, Optional

from everysk.api import File
from everysk.api.utils import EveryskList
from mcp.server.fastmcp import FastMCP

from everysk_mcp.fields import definitions as fields
from everysk_mcp.utils import from_base64_to_utf8, from_utf8_to_base64


###############################################################################
# Implementation
###############################################################################
def register_file_tools(mcp: FastMCP) -> None:
    """Register all file tools with the given MCP instance."""

    @mcp.tool(
        description="Returns a list of files previously uploaded. The files are returned in sorted order, with the most recent file appearing first."
    )
    def file_list(
        query: Optional[str] = fields.query,
        workspace: Optional[str] = fields.workspace,
        page_size: Optional[int] = fields.page_size,
    ) -> EveryskList[File]:
        result = File.list(query=query, workspace=workspace, page_size=page_size)
        return result

    @mcp.tool(
        description="Retrieves the details of an existing file by supplying the file's id."
    )
    def file_retrieve(
        id: str = fields.id_, workspace: Optional[str] = fields.workspace
    ) -> File:
        file = File.retrieve(id=id, workspace=workspace)
        file["data"] = from_base64_to_utf8(file["data"]) if file["data"] else None
        return file

    @mcp.tool(description="Creates and then returns the new file.")
    def file_create(
        name: str = fields.name,
        description: str = fields.description,
        tags: Optional[List[str]] = fields.tags,
        data: str = fields.data_file,
        with_data: Optional[bool] = fields.with_data,
        content_type: Optional[
            Literal[
                "image/svg+xml",
                "image/bmp",
                "image/jpeg",
                "image/png",
                "image/gif",
                "application/xml",
                "text/xml",
                "application/javascript",
                "application/json",
                "text/plain",
                "text/csv",
                "application/csv",
                "text/x-comma-separated-values",
                "text/comma-separated-values",
                "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
                "application/vnd.ms-excel",
                "application/pdf",
                "application/zip",
                "application/x-zip-compressed",
                "application/octet-stream",
            ]
        ] = fields.content_type,
        workspace: Optional[str] = fields.workspace,
    ) -> File:
        result = File.create(
            name=name,
            description=description,
            tags=tags,
            data=from_utf8_to_base64(data),
            with_data=with_data,
            content_type=content_type,
            workspace=workspace,
        )
        return result



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from mcp.server.fastmcp import FastMCP

from everysk_mcp.prompts.templates import register_prompts
from everysk_mcp.resources.static import register_static_resources
from everysk_mcp.resources.templates import register_template_resources
from everysk_mcp.tools.calculation import register_calculation_tools
from everysk_mcp.tools.datastore import register_datastore_tools
from everysk_mcp.tools.file import register_file_tools
from everysk_mcp.tools.portfolio import register_portfolio_tools
from everysk_mcp.tools.report import register_report_tools
from everysk_mcp.tools.report_template import register_report_template_tools
from everysk_mcp.tools.workflow import register_workflow_tools
from everysk_mcp.tools.workflow_execution import register_workflow_execution_tools
from everysk_mcp.tools.workspace import register_workspace_tools

###############################################################################
# Implementation
###############################################################################

mcp = FastMCP(
    "Everysk MCP Server",
    instructions="This is a very workflow-centric assistant, so always try to use an existing workflow to complete tasks and to answer questions. To this end, the first action will always be to immediately list the workflows available in the workspace and keep this information ready to use. Whenever possible, results will be rendered on the screen. Only when asked code will be generated for reports, tables and other deliverables. Whenever possible, symbols will be used in the output to the user instead of sec_ids.",
)

# Register all prompts
register_prompts(mcp)

# Register all resources
register_static_resources(mcp)
register_template_resources(mcp)

# Register all tools
register_calculation_tools(mcp)
register_datastore_tools(mcp)
register_file_tools(mcp)
register_portfolio_tools(mcp)
register_report_template_tools(mcp)
register_report_tools(mcp)
register_workflow_execution_tools(mcp)
register_workflow_tools(mcp)
register_workspace_tools(mcp)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.api import WorkflowExecution
from mcp.server.fastmcp import FastMCP

from everysk_mcp.fields import definitions as fields


###############################################################################
# Implementation
###############################################################################
def register_workflow_execution_tools(mcp: FastMCP) -> None:
    """Register all workflow execution tools with the given MCP instance."""

    @mcp.tool(
        description="Retrieves the details of an existing workflow execution by supplying the execution's id."
    )
    def workflow_execution_retrieve(
        workflow_id: str = fields.id_, workflow_execution_id: str = fields.id_
    ) -> WorkflowExecution:
        execution = WorkflowExecution.retrieve(
            workflow_id=workflow_id, workflow_execution_id=workflow_execution_id
        )
        return execution



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import List, Literal, Optional

from everysk.api import Calculation
from mcp.server.fastmcp import FastMCP

from everysk_mcp.blueprints.securities import Security
from everysk_mcp.fields import definitions as fields


###############################################################################
# Implementation
###############################################################################
def validate_input(
    portfolio_id: Optional[str], securities: Optional[List[Security]]
) -> Optional[List[dict]]:
    """
    Validates mutually exclusive parameters: `portfolio_id` or `securities`.

    - Raises a ValueError if both or neither are provided.
    - Validates that `securities`, if provided, is a list of `Security` instances.
    - Returns the serialized list of securities if valid.

    Args:
        portfolio_id (Optional[str]): The ID of the portfolio.
        securities (Optional[List[Security]]): List of Security objects.

    Returns:
        Optional[List[dict]]: A list of serialized Security dicts if `securities` is provided.
    """
    if portfolio_id is not None and securities is not None:
        raise ValueError(
            "You must provide either portfolio_id or securities, but not both."
        )

    if portfolio_id is None and securities is None:
        raise ValueError("You must provide either portfolio_id or securities.")

    if securities is not None:
        if not isinstance(securities, list):
            raise TypeError("Securities must be a list of Security objects.")

        for security in securities:
            if not isinstance(security, Security):
                raise TypeError(
                    "Each security must be an instance of the Security class."
                )

        return [security.model_dump() for security in securities]

    return None


def register_calculation_tools(mcp: FastMCP) -> None:
    """Register all calculation tools with the given MCP instance."""

    @mcp.tool(
        description="The goal of Monte Carlo Risk Attribution is to calculate the contribution to the overall portfolio risk from each security. We use a measure called Marginal Contribution to Total Risk (MCTR) for Monte Carlo Risk Attribution."
    )
    def monte_carlo_risk_attribution(
        portfolio_id: Optional[str] = fields.id_,
        securities: Optional[List[Security]] = fields.security_list,
        date: Optional[str] = fields.date_portfolio,
        base_currency: Optional[str] = fields.base_currency,
        nlv: Optional[float] = fields.nlv,
        horizon: Optional[int] = fields.horizon,
        sampling: Optional[int] = fields.sampling,
        aggregation: Optional[
            Literal[
                "custom",
                "position",
                "country",
                "sector",
                "gics_sector",
                "market_capitalization",
                "liquidity",
                "implied_rating",
                "duration",
                "security_type",
                "security_type_refined",
                "dividend_yield",
                "exposure",
                "currency",
                "fixed_income",
                "total_esg",
            ]
        ] = fields.aggregation,
        projection: Optional[list[str]] = fields.projection,
        volatility_half_life: Optional[int] = fields.volatility_half_life,
        correlation_half_life: Optional[int] = fields.correlation_half_life,
        risk_measure: Optional[Literal["vol", "var", "cvar"]] = fields.risk_measure,
        filter: Optional[str] = fields.filter_,
    ) -> dict:
        securities = validate_input(portfolio_id, securities)

        result = Calculation.riskAttribution(
            portfolio_id=portfolio_id,
            securities=securities,
            date=date,
            base_currency=base_currency,
            nlv=nlv,
            horizon=horizon,
            sampling=sampling,
            aggregation=aggregation,
            projection=projection,
            volatility_half_life=volatility_half_life,
            correlation_half_life=correlation_half_life,
            risk_measure=risk_measure,
            filter=filter,
        )

        return result

    @mcp.tool(
        description="Calculates the delta-adjusted notional exposure of each security, converted to the base currency of the portfolio."
    )
    def exposure(
        portfolio_id: Optional[str] = fields.id_,
        securities: Optional[List[Security]] = fields.security_list,
        date: Optional[str] = fields.date_portfolio,
        base_currency: Optional[str] = fields.base_currency,
        nlv: Optional[float] = fields.nlv,
        sampling: Optional[int] = fields.sampling,
        aggregation: Optional[
            Literal[
                "custom",
                "position",
                "country",
                "sector",
                "gics_sector",
                "market_capitalization",
                "liquidity",
                "implied_rating",
                "duration",
                "security_type",
                "security_type_refined",
                "dividend_yield",
                "exposure",
                "currency",
                "fixed_income",
                "total_esg",
            ]
        ] = fields.aggregation,
        filter: Optional[str] = fields.filter_,
    ):
        securities = validate_input(portfolio_id, securities)

        result = Calculation.exposure(
            portfolio_id=portfolio_id,
            securities=securities,
            date=date,
            base_currency=base_currency,
            nlv=nlv,
            sampling=sampling,
            aggregation=aggregation,
            filter=filter,
        )

        return result

    @mcp.tool(
        description="Calculates the overall portfolio properties with a single API call. Sensitivities, exposures and risk are aggregated from individual securities to a portfolio level."
    )
    def properties(
        portfolio_id: Optional[str] = fields.id_,
        securities: Optional[List[Security]] = fields.security_list,
        date: Optional[str] = fields.date_portfolio,
        base_currency: Optional[str] = fields.base_currency,
        nlv: Optional[float] = fields.nlv,
        horizon: Optional[int] = fields.horizon,
        sampling: Optional[int] = fields.sampling,
        aggregation: Optional[
            Literal[
                "custom",
                "position",
                "country",
                "sector",
                "gics_sector",
                "market_capitalization",
                "liquidity",
                "implied_rating",
                "duration",
                "security_type",
                "security_type_refined",
                "dividend_yield",
                "exposure",
                "currency",
                "fixed_income",
                "total_esg",
            ]
        ] = fields.aggregation,
        projection: Optional[list[str]] = fields.projection,
        volatility_half_life: Optional[int] = fields.volatility_half_life,
        correlation_half_life: Optional[int] = fields.correlation_half_life,
        confidence: Optional[str] = fields.confidence,
    ):
        securities = validate_input(portfolio_id, securities)

        result = Calculation.properties(
            portfolio_id=portfolio_id,
            securities=securities,
            date=date,
            base_currency=base_currency,
            nlv=nlv,
            horizon=horizon,
            sampling=sampling,
            aggregation=aggregation,
            projection=projection,
            volatility_half_life=volatility_half_life,
            correlation_half_life=correlation_half_life,
            confidence=confidence,
        )

        return result

    @mcp.tool(
        description='Calculates greeks for options and various fixed income sensitivities for fixed income securities. Values can be unitized when "compute_notional" is False, or weighted by the notional exposure of the security when "compute_notional" is True.'
    )
    def sensitivity(
        portfolio_id: Optional[str] = fields.id_,
        securities: Optional[List[Security]] = fields.security_list,
        date: Optional[str] = fields.date_portfolio,
        base_currency: Optional[str] = fields.base_currency,
        sensitivity_type: Literal[
            "delta",
            "gamma",
            "theta",
            "vega",
            "duration",
            "effective_duration",
            "macaulay_duration",
            "modified_duration",
            "cs01",
            "dv01",
            "average_life",
            "rolled_average_life",
            "volatility",
        ] = fields.sensitivity_type,
        compute_notional: Optional[bool] = fields.compute_notional,
    ):
        securities = validate_input(portfolio_id, securities)

        result = Calculation.sensitivity(
            portfolio_id=portfolio_id,
            securities=securities,
            date=date,
            base_currency=base_currency,
            sensitivity_type=sensitivity_type,
            compute_notional=compute_notional,
        )

        return result

    @mcp.tool(
        description="Tracking error is the divergence between the price behavior of a position or a portfolio and the price behavior of a benchmark. This is often in the context of a hedge fund, mutual fund, or exchange-traded fund (ETF) that did not work as effectively as intended, creating an unexpected profit or loss."
    )
    def tracking_errors(
        portfolio1: str = fields.id_,
        portfolio2: str = fields.id_,
        weekly: Optional[bool] = fields.weekly,
        EWMA: Optional[bool] = fields.ewma,
        exponential_decay: Optional[float] = fields.exponential_decay,
        nDays: Optional[int] = fields.ndays,
        nWeeks: Optional[int] = fields.nweeks,
    ):
        result = Calculation.marginalTrackingError(
            portfolio1=portfolio1,
            portfolio2=portfolio2,
            weekly=weekly,
            EWMA=EWMA,
            exponential_decay=exponential_decay,
            nDays=nDays,
            nWeeks=nWeeks,
        )

        return result

    @mcp.tool(
        description="Calculates the magnitude of a security's reaction to changes in underlying factors, most often in terms of its price response to other factors."
    )
    def parametric_risk_attribution(
        id: str = fields.id_,
        projection: List[Security] = fields.projection,
        portfolio_id: Optional[str] = fields.id_exclusive,
        securities: Optional[List[Security]] = fields.security_list_exclusive,
        use_cashflow: Optional[bool] = fields.use_cashflow,
        sampling: Optional[int] = fields.sampling,
        confidence: Optional[str] = fields.confidence,
        historical_days: Optional[int] = fields.historical_days,
        exponential_decay: Optional[float] = fields.exponential_decay,
    ):
        securities = validate_input(portfolio_id, securities)

        result = Calculation.parametricRiskAttribution(
            id=id,
            projection=projection,
            portfolio_id=portfolio_id,
            securities=securities,
            use_cashflow=use_cashflow,
            sampling=sampling,
            confidence=confidence,
            historical_days=historical_days,
            exponential_decay=exponential_decay,
        )

        return result

    @mcp.tool(
        description="Stress Test calculates the expected behavior of the portfolio under different scenarios, including extreme events. Everysk API calculates the probabilities of extreme events happening and factors those probabilities into the calculations. For example: it takes into account that correlations and volatility tend to increase in periods of market distress, resulting in large price oscillations for securities.This API call returns 3 main measures for the specific scenario requested, namely: Conditional value at risk at the left tail of the distribution (CVaR-): The average of the worst (1-confidence) percentile outcomes from the forward looking P&L distribution of the portfolio. Expected profit and loss (EV): The average of the full forward looking P&L distribution of the portfolio. Conditional value at risk at the right tail of the distribution (CVaR+): The average of the best (1-confidence) percentile outcomes from the forward looking P&L distribution of the portfolio."
    )
    def stress_test(
        portfolio_id: Optional[str] = fields.id_exclusive,
        securities: Optional[List[Security]] = fields.security_list_exclusive,
        date: Optional[str] = fields.date_portfolio,
        base_currency: Optional[str] = fields.base_currency,
        nlv: Optional[float] = fields.nlv,
        horizon: Optional[int] = fields.horizon,
        sampling: Optional[int] = fields.sampling,
        aggregation: Optional[
            Literal[
                "custom",
                "position",
                "country",
                "sector",
                "gics_sector",
                "market_capitalization",
                "liquidity",
                "implied_rating",
                "duration",
                "security_type",
                "security_type_refined",
                "dividend_yield",
                "exposure",
                "currency",
                "fixed_income",
                "total_esg",
            ]
        ] = fields.aggregation,
        projection: Optional[list[str]] = fields.projection,
        volatility_half_life: Optional[int] = fields.volatility_half_life,
        correlation_half_life: Optional[int] = fields.correlation_half_life,
        shock: Optional[str] = fields.shock,
        magnitude: str = fields.magnitude,
        confidence: Optional[str] = fields.confidence,
        filter: Optional[str] = fields.filter_,
    ):
        securities = validate_input(portfolio_id, securities)

        result = Calculation.stressTest(
            portfolio_id=portfolio_id,
            securities=securities,
            date=date,
            base_currency=base_currency,
            nlv=nlv,
            horizon=horizon,
            sampling=sampling,
            aggregation=aggregation,
            projection=projection,
            volatility_half_life=volatility_half_life,
            correlation_half_life=correlation_half_life,
            shock=shock,
            magnitude=magnitude,
            confidence=confidence,
            filter=filter,
        )

        return result

    # @mcp.tool(description="Calculates the number of days required to unwind a position in the portfolio, considering the liquidity of each asset. This calculation can help in assessing the market risk associated with the portfolio's current composition.")
    # def days_to_unwind():
    #     ...

    # @mcp.tool(description='Calculates sensitivity for each underlying stress scenario. `securities` parameter is not required if `portfolio_id` is being provided.')
    # def underlying_stress_sensitivity():
    #     ...

    # @mcp.tool(description='Retrieves fundamental data.')
    # def fundamentals():
    #     ...

    # @mcp.tool(description='Retrieves backtest statistics by executing a backtest based on the provided parameters. This method internally calls a function to compute backtest statistics and returns a detailed result dict.')
    # def backtestStatistics():
    #     ...

    # NOT TO BE IMPLEMENTED. WILL BE RENAMED TO portfolio_pricer
    # def bond_pricer():
    #     ...



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Any, Dict, Optional

from everysk.api import Workflow, WorkflowExecution
from everysk.api.utils import EveryskList
from mcp.server.fastmcp import FastMCP

from everysk_mcp.fields import definitions as fields


###############################################################################
# Implementation
###############################################################################
def register_workflow_tools(mcp: FastMCP) -> None:
    """Register all workflow tools with the given MCP instance."""

    @mcp.tool(
        description="Returns a list of workflows you've previously created. The workflows are returned in sorted order, with the most recent workflow appearing first."
    )
    def workflow_list(
        query: Optional[str] = fields.query,
        workspace: Optional[str] = fields.workspace,
        page_size: Optional[int] = fields.page_size,
    ) -> EveryskList[Workflow]:
        result = Workflow.list(query=query, workspace=workspace, page_size=page_size)
        return result

    @mcp.tool(
        description="Retrieves the details of an existing workflow by supplying the workflow's id."
    )
    def workflow_retrieve(
        id: str = fields.id_, workspace: Optional[str] = fields.workspace
    ) -> Workflow:
        result = Workflow.retrieve(id=id, workspace=workspace)
        return result

    @mcp.tool(
        description="Provides information about the required parameters to run a workflow and details on its output. It should always be called before running a workflow using the `workflow_run` tool. This tool retrieves the workflow details and prepares the context for running it, so I should not use `workflow_retrieve` if I am going to use this tool."
    )
    def workflow_prepare_run(
        id: str = fields.id_,
        workspace: Optional[str] = fields.workspace,
    ) -> str:
        wf = workflow_retrieve(id=id, workspace=workspace)

        for worker in wf.workers:
            if worker["id"] == wf.starter_worker_id:
                starter_context = str(worker["script_outputs"])
            elif worker["id"] == wf.ender_worker_id:
                ender_context = str(worker["script_outputs"])

        return (
            f"The workflow `{wf.name}` accepts the following parameters:\n\n"
            f"{starter_context}"
            f"The workflow `{wf.name}` will return the following outputs:\n\n"
            f"{ender_context}"
            f"You can now run the workflow using the `workflow_run` tool with the required parameters."
        )

    @mcp.tool(
        description="Runs a specific workflow through the api. It should be called after the `workflow_prepare_run` tool to ensure that the workflow is ready to be executed with the provided parameters."
    )
    def workflow_run(
        id: str = fields.id_,
        workspace: Optional[str] = fields.workspace,
        parameters: Dict[str, Any] = fields.parameters,
        synchronous: Optional[bool] = fields.synchronous,
    ) -> WorkflowExecution:
        execution = Workflow.run(
            id=id, workspace=workspace, parameters=parameters, synchronous=synchronous
        )
        return execution
